/* Placed in the public domain 2001 by Sam Trenholme */

/* Simple testing DNS resolving client.  This sends out simple
   queries to DNS servers and prints out to stdout the replies
   it receives.
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "../MaraDns.h"
/* All of the labels */
#include "askmara_labels_en.h"

int harderror(char *msg) {
    printf("%s%s%s",L_HARD_ERROR,msg,L_NEWLINE);
    exit(1);
    }

main(int argc, char **argv) {
    char *server_address = NULL;
    struct sockaddr_in dns_udp, server;
    int len_inet; /* Length */
    int s; /* Socket */
    int q; /* Used for binding, etc */
    js_string *outdata; /* Outgoing data */
    js_string *indata, *uindata; /* Incoming data (uncompressed version) */
    q_header header; /* header data */
    q_question question;
    int counter,place,count;
    fd_set rx_set; /* Using select() because if its timeout option */ 
    int maxd;      /* select() */
    struct timeval tv;  /* select() */
    int n; /* Select() return value */

    /* Determine what the query string is */
    if(argc < 2 || argc > 3) 
        harderror(L_USAGE); 

    /* Determine what IP address to bind to */
    if(argc == 3) 
        server_address = argv[2];
    else 
        server_address = "127.0.0.3";

    if((indata = js_create(512,1)) == 0)
       harderror(L_JS_CREATE_INDATA);
    if((uindata = js_create(2048,1)) == 0)
       harderror(L_JS_CREATE_UINDATA);
    if((outdata = js_create(512,1)) == 0)
       harderror(L_JS_CREATE_OUTDATA);
    if((question.qname = js_create(512,1)) == 0)
       harderror(L_JS_CREATE_QNAME); 
    /* Create a UDP client socket */
    if((s = socket(AF_INET,SOCK_DGRAM,0)) == -1)
        harderror(L_SOCKET);

    /* Create a socket address to use with sendto() */
    memset(&dns_udp,0,sizeof(dns_udp));
    dns_udp.sin_family = AF_INET;
    dns_udp.sin_port = htons(53);
    if((dns_udp.sin_addr.s_addr = inet_addr(server_address)) == INADDR_NONE)
        harderror(L_MAL_IP);

    len_inet = sizeof(dns_udp);

    /* Format a DNS request */
    /* DNS header -> using dig defaults */
    header.id = 6;
    header.qr = 0;
    header.opcode = 0;
    header.aa = 0;
    header.tc = 0;
    header.rd = 1; /* Recursion desired */
    header.ra = 0;
    header.z = 0;
    header.rcode = 0;
    header.qdcount = 1;
    header.ancount = 0;
    header.nscount = 0;
    header.arcount = 0;

    /* Create a DNS question , and put it in raw UDP format */
    /* DNS question -> looking up an A record... */
    question.qclass = 1; /* ...on the internet */
    if(js_qstr2js(question.qname,argv[1]) == JS_ERROR)
        harderror(L_INVALID_Q); /* Invalid query */

    /* Make 'Aexample.com.' raw UDP data */
    
    place = hname_2rfc1035(question.qname); 
    if(place == JS_ERROR)
        harderror(L_INVALID_DQ); /* Invalid form of domain query */
    question.qtype = place; 

    /* Make a string containing the RAW UDP query */
    make_hdr(&header,outdata);
    make_question(&question,outdata); 

    /* Send out a DNS request */
    if(sendto(s,outdata->string,outdata->unit_count,0,
       (struct sockaddr *)&dns_udp,len_inet) < 0) 
        harderror(L_UDP_NOSEND); /* Unable to send UDP packet */

    /* Wait for a reply from the DNS server */
    FD_ZERO(&rx_set);
    FD_SET(s,&rx_set);
    maxd = s + 1;
    tv.tv_sec = 2;
    tv.tv_usec = 0;
    n = select(maxd,&rx_set,NULL,NULL,&tv);
    if(n == -1)  /* select error */
        harderror("Select() failed");
    if(n == 0) /* Timeout */
        harderror("Timeout");
    if((count = recvfrom(s,indata->string,indata->max_count,0,
	                (struct sockaddr *)&dns_udp,&len_inet)) < 0)
        harderror(L_DNS_R_ERROR); /* Problem getting DNS server response */
 
    indata->unit_count = count;

    decompress_data(indata,uindata);

    /* 
    js_show_stdout(indata);
    printf("%s",L_NEWLINE);
    js_show_stdout(uindata); 
    */

    /* Print out the reply -- cut 'n paste from rtest.c */
    if(read_hdr(uindata,&header) == JS_ERROR) 
        harderror(L_INHEADER_CONV); /* Problem converting inheader */

    printf("%s%s",L_SERVER_REPLY,L_NEWLINE); /* Server reply */
    printf("%s%d%s",L_QUERY_ID,header.id,L_NEWLINE); /* Query ID */
    printf("%s%d%s",L_QUERY_TYPE,header.qr,L_NEWLINE); /* Query type */
    printf("%s%d%s",L_OPCODE,header.opcode,L_NEWLINE); /* Opcode */
    printf("%s%d%s",L_AUTHORITATIVE,header.aa,L_NEWLINE); /* Auth. */
    printf("%s%d%s",L_TRUNCATED,header.tc,L_NEWLINE); /* Truncated */
    printf("%s%d%s",L_RD,header.rd,L_NEWLINE); /* Recursion desired */
    printf("%s%d%s",L_RA,header.ra,L_NEWLINE); /* Recursion available */
    printf("%s%d%s",L_Z_DATA,header.z,L_NEWLINE); /* Z-Data */
    printf("%s%d%s",L_RC,header.rcode,L_NEWLINE); /* Result code */
    printf("%s%d%s",L_QDCOUNT,header.qdcount,L_NEWLINE); /* Num. questions */
    printf("%s%d%s",L_ANCOUNT,header.ancount,L_NEWLINE); /* Num. answers */
    printf("%s%d%s",L_NSCOUNT,header.nscount,L_NEWLINE); /* Num. NS RRs */
    printf("%s%d%s",L_ARCOUNT,header.arcount,L_NEWLINE); /* Num. Additional */
    printf("%s",L_NEWLINE);

    /* Get all of the questions in the reply */
    place = 12;
    for(counter = 0;counter < header.qdcount;counter++) {
        count = read_question(uindata,&question,place);     
        if(count < 0) /* Error handling */
	    harderror(L_QERROR); /* Error reading question */
        place += count;
	printf("%s",L_QNAME); /* Question name */
	hname_translate(question.qname,question.qtype);
	js_show_stdout(question.qname);
	printf("%s%s%d%s",L_NEWLINE,L_QTYPE,question.qtype,L_NEWLINE); 
	printf("%s%d%s",L_QCLASS,question.qclass,L_NEWLINE); /* Ques. Class */
	}

    /* Get all of the answers in the reply */
    /* Get all of the an replies */
    printf("%s%s%s",L_NEWLINE,L_ANREP,L_NEWLINE);
    for(counter = 0; counter < header.ancount; counter++) 
        out_answer(uindata,&place);

    /* Get all of the ns replies */
    printf("%s%s%s",L_NEWLINE,L_NSREP,L_NEWLINE);
    for(counter = 0; counter < header.nscount; counter++) 
        out_answer(uindata,&place);

    /* Get all of the ar replies */
    printf("%s%s%s",L_NEWLINE,L_ARREP,L_NEWLINE);
    for(counter = 0; counter < header.arcount; counter++) 
        out_answer(uindata,&place);

    }

/* Parse the answer part of a DNS query.
   input: pointer to js_string with uncompressed UDP data
          pointer to where we are in that string
   output: outputs to stdout the data in that string
*/

int out_answer(js_string *uindata,int *place) {

        q_rr rr_hdr;
        rr_soa soa;
	rr_mx mx;
        int count;

    /* Create the needed strings */
    if((rr_hdr.name = js_create(512,1)) == 0)
       harderror(L_HNAME_OA); /* js_create with hname in oa */
    if((soa.mname = js_create(512,1)) == 0)
       harderror(L_C_MNAME); /* js_create with soa.mname in oa */
    if((soa.rname = js_create(512,1)) == 0)
       harderror(L_C_RNAME); /* js_create with soa.rname in oa */
    if((mx.exchange = js_create(512,1)) == 0)
       harderror(L_C_MXEXC); /* js_create with mx.exchange in oa */

        /* Start printing out the data header */
        count = read_rr_h(uindata,&rr_hdr,*place);
	if(count < 0)
	    harderror(L_RR_ERR); /* Error reading rr in AN section */
        *place += count;
        hname_translate(rr_hdr.name,rr_hdr.type);
        printf("%s",L_RNAME); /* Record name */
	js_show_stdout(rr_hdr.name);
	printf("\nRecord type: %d\n",rr_hdr.type);
	printf("Record class: %d\n",rr_hdr.class);
	printf("Record TTL: %u\n",rr_hdr.ttl);
	printf("Record length: %d\n",rr_hdr.rdlength);
	printf("%s",L_NEWLINE);
        if(rr_hdr.type == RR_SOA) {
	    if(read_soa(uindata,&soa,*place) == JS_ERROR)
                harderror(L_READSOA);
            hname_translate(soa.mname,RR_SOA);
            hname_translate(soa.rname,RR_MX);
	    printf("%s",L_S_MNAME); /* SOA mname */
	    js_show_stdout(soa.mname);
	    printf("%s%s",L_NEWLINE,L_S_RNAME); /* SOA rname */
	    js_show_stdout(soa.rname);
	    printf("%s%s%u%s",L_NEWLINE,L_S_SERIAL,soa.serial,L_NEWLINE);
	    printf("%s%d%s",L_S_REFRESH,soa.refresh,L_NEWLINE);
	    printf("%s%d%s",L_S_RETRY,soa.retry,L_NEWLINE);
	    printf("%s%d%s",L_S_EXPIRE,soa.expire,L_NEWLINE);
	    printf("%s%u%s",L_S_MINIMUM,soa.minimum,L_NEWLINE);
            } 
        else if(rr_hdr.type == RR_MX) {
	    if(uindata->unit_count < *place + 2)
	        return JS_ERROR;
            mx.preference = ((*(uindata->string + *place) & 0xff) << 8) |
                             (*(uindata->string + *place + 1) & 0xff);
	    read_ns(uindata,mx.exchange,*place + 2);
	    printf("%s%d%s",L_M_PREF,mx.preference,L_NEWLINE); /* MX pref. */
	    printf("%s",L_M_EX); /* MX exchange */
            hname_translate(mx.exchange,RR_MX);
	    js_show_stdout(mx.exchange);
	    printf("%s%s",L_NEWLINE,L_NEWLINE);
	    }
        else if(rr_hdr.type == RR_NS || rr_hdr.type == RR_CNAME || 
	        rr_hdr.type == RR_PTR) {
	    printf("%s",L_DATA); /* Data: */
	    read_ns(uindata,mx.exchange,*place);
            hname_translate(mx.exchange,rr_hdr.type);
	    js_show_stdout(mx.exchange);
	    printf("%s%s",L_NEWLINE,L_NEWLINE);
	    }
        else if(rr_hdr.type == RR_A) {
	    if(uindata->unit_count < *place + 4)
	        return JS_ERROR;
            /* Display the IP of the data */
	    printf("%s%d%s%d%s%d%s%d%s%s",L_IP,*(uindata->string + *place),
	           L_DOT,*(uindata->string + *place + 1),
		   L_DOT,*(uindata->string + *place + 2),
		   L_DOT,*(uindata->string + *place + 3),L_NEWLINE,L_NEWLINE);
            }
        else if(rr_hdr.type = RR_TXT) {
	    printf("%s",L_TXT); /* Text String */
            read_txt(uindata,mx.exchange,*place); 
	    js_show_stdout(mx.exchange);
	    printf("%s%s",L_NEWLINE,L_NEWLINE);
	    }
        else 
            printf("%s%d%s%s",L_UNSUP,rr_hdr.type,L_NEWLINE,L_NEWLINE);
	    
	*place += rr_hdr.rdlength; /* To do: read the RD data itself for
	                              all MaraDNS supported RRs */
	}
